/* 4567890123456789012345678901234567890123456789012345678901234567890123456789
**
** File      cgi.c
**
** Desc      A CGI library for C programmers. Automatically handles GET, POST
**           and ISINDEX requests, with lots of useful routines to simplify
**           CGI development. Applications can even be debugged from the
**           commandline!
**
** Usage     See http://www.jemtek.com.au/jemtek/cgi/lib
**
** Limits    Cannot yet handle form file uploads.
**
** Copyright 1996 JemTek Pty Ltd, Jerome Munchenberg
**           http://www.jemtek.com.au/jemtek
**           jerome@jemtek.com.au
**
** History   Ver  Date     Author Modification
**           1.0  06-03-96 JEM    First Draft
**                29-07-96 JEM    Modified error messages in main
**	          31-08-96 JEM    Swapped LFs for NLs in http headers
**	          06-12-96 JEM    Fixed ambiguous pointers in cgiDecodeUrl
**
******************************************************************************/

/* Includes 
******************************************************************************/

#include <stdio.h>
#include <stdlib.h> /* getenv */
#include <stdarg.h> /* vfprintf */
#include <string.h> /* str... */
#include <ctype.h>  /* is... */

#include "cgi.h"

/* Local global variables
******************************************************************************/

static int    cgiHeaderFlag = 0;
static int    cgiHtmlFlag   = 0;

static int    cgiArgc  = 0;
static char **cgiArgv  = NULL;
static char  *cgiStdin = NULL;

typedef struct {
  char *var;
  char *val;
} pair_t;

typedef struct {
  int     size;
  pair_t *pair;
} form_t;

form_t form; 

/* Local function declaration
******************************************************************************/

static char *cgiDebugGetStr(FILE *fp);
static int   cgiDebugGetInt(FILE *fp);
static int   cgiDebugEnvLoad(char *file);
static int   cgiDebugEnvSave(char *file);

static char *cgiDecodeUrl(char *data);

static int   cgiFormAdd(char *var, char *val);

static char *cgiGetEnv(char *var);
static char *cgiGetStdin(int len);

static char  cgiHexAsc(char x1, char x2);
static int   cgiHexInt(char x);

static int   cgiParseArgs(char **argv, int argc);
static int   cgiParseInput(char *input, int len);

/* Library functions
******************************************************************************/

/******************************************************************************
**
** Function  cgiDisplayEnv
**
** Purpose   Display CGI specific environment variables stored in cgiEnv struc
**
** History   Date     Author Modification
**           04-03-96 JEM    First Draft
**
******************************************************************************/

void cgiDisplayEnv(void)
{
  if (!cgiHeaderFlag)
    return;

  cgiPrintf("<H2>CGI Global Structure</H2>");
  cgiPrintf("<PRE>");
  cgiPrintf("%-23s [%s]", "cgiEnv.authType",         cgiEnv.authType);
  cgiPrintf("%-23s [%i]", "cgiEnv.contentLength",    cgiEnv.contentLength);
  cgiPrintf("%-23s [%s]", "cgiEnv.contentType",      cgiEnv.contentType);
  cgiPrintf("%-23s [%s]", "cgiEnv.gatewayInterface", cgiEnv.gatewayInterface);
  cgiPrintf("%-23s [%s]", "cgiEnv.httpAccept",       cgiEnv.httpAccept);
  cgiPrintf("%-23s [%s]", "cgiEnv.httpHost",         cgiEnv.httpHost);
  cgiPrintf("%-23s [%s]", "cgiEnv.httpReferer",      cgiEnv.httpReferer);
  cgiPrintf("%-23s [%s]", "cgiEnv.httpUserAgent",    cgiEnv.httpUserAgent);
  cgiPrintf("%-23s [%s]", "cgiEnv.pathInfo",         cgiEnv.pathInfo);
  cgiPrintf("%-23s [%s]", "cgiEnv.pathTranslated",   cgiEnv.pathTranslated);
  cgiPrintf("%-23s [%s]", "cgiEnv.queryString",      cgiEnv.queryString);
  cgiPrintf("%-23s [%s]", "cgiEnv.remoteAddr",       cgiEnv.remoteAddr);
  cgiPrintf("%-23s [%s]", "cgiEnv.remoteHost",       cgiEnv.remoteHost);
  cgiPrintf("%-23s [%s]", "cgiEnv.remoteIdent",      cgiEnv.remoteIdent);
  cgiPrintf("%-23s [%s]", "cgiEnv.remoteUser",       cgiEnv.remoteUser);
  cgiPrintf("%-23s [%s]", "cgiEnv.requestMethod",    cgiEnv.requestMethod);
  cgiPrintf("%-23s [%s]", "cgiEnv.scriptName",       cgiEnv.scriptName);
  cgiPrintf("%-23s [%s]", "cgiEnv.serverName",       cgiEnv.serverName);
  cgiPrintf("%-23s [%s]", "cgiEnv.serverPort",       cgiEnv.serverPort);
  cgiPrintf("%-23s [%s]", "cgiEnv.serverProtocol",   cgiEnv.serverProtocol);
  cgiPrintf("%-23s [%s]", "cgiEnv.serverSoftware",   cgiEnv.serverSoftware);
  cgiPrintf("</PRE>");
}

/******************************************************************************
**
** Function  cgiDisplayPairs
**
** Purpose   Display var=val pairs from form
**
** History   Date     Author Modification
**           06-03-96 JEM    First Draft
**
******************************************************************************/

void cgiDisplayPairs()
{
  int cnt;
  
  if (!cgiHeaderFlag)
    return;

  cgiPrintf("<H2>Form Pairs</H2>");
  cgiPrintf("<PRE>");
  if(!form.size)
    cgiPrintf("(no data)");
  else
    for(cnt=0; cnt<form.size; cnt++)
      cgiPrintf("%2i %-12s [%s]", cnt, form.pair[cnt].var, form.pair[cnt].val);
  cgiPrintf("</PRE>");
}

/******************************************************************************
**
** Function  cgiEncodeSys
**
** Purpose   Escape shell metacharacters in a string (for calls to popen etc)
**
** History   Date     Author Modification
**           15-03-96 JEM    First Draft
**
******************************************************************************/

char *cgiEncodeSys(char *data)
{
  char *i, *j, *code;
  int   inc;

  for (inc=0, i=data; *i!='\0'; i++)
    if (strchr("&;`'|*?~<>^()[]{}$\"\n\\", *i))
      ++inc;
  
  if (!(code = (char*)malloc(strlen(data)+inc+1)))
    return NULL;
  
  for (j=code, i=data; *i!='\0'; i++, j++)
    {
      if (strchr("&;`'|*?~<>^()[]{}$\"\n\\", *i))
	*j++ = '\\';
      *j = *i;
    }
  *j = '\0';
  
  return code;
}

/******************************************************************************
**
** Function  cgiEncodeUrl
**
** Purpose   URL-encode a string (Convert space to plus, non-ascii to hex)
**
** History   Date     Author Modification
**           08-03-96 JEM    First Draft
**
******************************************************************************/

char *cgiEncodeUrl(char *data)
{
  char *hex = "0123456789ABCDEF";
  char *i, *j, *code;
  int   inc;

  for (inc=0, i=data; *i!='\0'; i++)
    if (!isalnum(*i))
      inc += 2;
  
  if (!(code = (char*)malloc(strlen(data)+inc+1)))
    return NULL;
  
  for (j=code, i=data; *i!='\0'; i++, j++)
    {
      if (*i == ' ')
	*j = '+';
      else
      if (!isalnum(*i))
	{
	  *j++ = '%';
	  *j++ = hex[*i/16];
	  *j   = hex[*i%16];
	}
      else
	*j = *i;
    }
  *j = '\0';
  
  return code;
}

/******************************************************************************
**
** Function  cgiFormCmpVal
**
** Purpose   Compares the value of a variable with another value
**
** History   Date     Author Modification
**           18-03-96 JEM    First Draft
**
******************************************************************************/

int cgiFormCmpVal(char *var, char *val)
{
  char **vals;
  int    cnt, len;
  
  len = cgiFormGetValMulti(var, &vals);

  for (cnt=0; cnt<len; cnt++)
    if (!strcmp(val, vals[cnt]))
      return 1;

  return 0;
}

/******************************************************************************
**
** Function  cgiFormCmpValMulti
**
** Purpose   Compares the value of a variable with an array of values
**
** History   Date     Author Modification
**           18-03-96 JEM    First Draft
**
******************************************************************************/

int cgiFormCmpValMulti(char *var, char len, char **vals)
{
  char **fvals;
  int    cnt, fcnt, flen;

  flen = cgiFormGetValMulti(var, &fvals);

  for (fcnt=0; fcnt<flen; fcnt++)
    for (cnt=0; cnt<len; cnt++)
      if (!strcmp(vals[cnt], fvals[fcnt]))
	return 1;

  return 0;
}

/******************************************************************************
**
** Function  cgiFormCntVar
**
** Purpose   Test for a variable
**
** History   Date     Author Modification
**           04-03-96 JEM    First Draft
**
******************************************************************************/

int cgiFormCntVar(char *var)
{
  int cnt, len;

  for(len=cnt=0; cnt<form.size; cnt++)
    if (!strcmp(var, form.pair[cnt].var))
      ++len;

  return len;
}

/******************************************************************************
**
** Function  cgiFormGetVal
**
** Purpose   Return the value of the given variable
**
** History   Date     Author Modification
**           04-03-96 JEM    First Draft
**
******************************************************************************/

char *cgiFormGetVal(char *var)
{
  int cnt;

  for(cnt=0; cnt<form.size; cnt++)
    if (!strcmp(var, form.pair[cnt].var))
      return strdup(form.pair[cnt].val);

  return NULL;
}

/******************************************************************************
**
** Function  cgiFormGetValMulti
**
** Purpose   Return multiple values of the given variable
**
** History   Date     Author Modification
**           04-03-96 JEM    First Draft
**
******************************************************************************/

int cgiFormGetValMulti(char *var, char ***vals)
{
  int len, cnt;

  *vals = NULL;

  for(len=cnt=0; cnt<form.size; cnt++)
    if (!strcmp(var, form.pair[cnt].var))
      {
 	if (!(*vals = (char**)realloc(*vals, (len+1)*sizeof(char*))))
	    return 0;
 	if (!((*vals)[len++] = strdup(form.pair[cnt].val)))
	    return 0;
      }

  return len;
}

/******************************************************************************
**
** Function  cgiHeaderContent
**
** Purpose   Indicate the MIME type of the document
**
** History   Date     Author Modification
**           04-03-96 JEM    First Draft
**
******************************************************************************/

int cgiHeaderContent(char *mimeType)
{
  if (cgiHeaderFlag++)
    return 0;

  printf("Content-Type: %s\n\n", mimeType);

  return 1;
}

/******************************************************************************
**
** Function  cgiHeaderLocation
**
** Purpose   Redirect the browser to a new resource
**
** History   Date     Author Modification
**           04-03-96 JEM    First Draft
**
******************************************************************************/

int cgiHeaderLocation(char *redirectUrl)
{
  if (cgiHeaderFlag++)
    return 0;
  
  printf("Location: %s\n\n", redirectUrl);

  return 1;
}

/******************************************************************************
**
** Function  cgiHeaderStatus
**
** Purpose   Return an error status
**
** History   Date     Author Modification
**           04-03-96 JEM    First Draft
**
******************************************************************************/

int cgiHeaderStatus(int statusCode, char *statusReason, char *statusMesg)
{
  if (cgiHeaderFlag++)
    return 0;
    
  printf("Content-type: text/html\n");
  printf("Status: %i %s\n\n", statusCode, statusReason);

  cgiPrintf("<HTML>");
  cgiPrintf("<HEAD>");
  cgiPrintf("<TITLE>%i %s</TITLE>", statusCode, statusReason);
  cgiPrintf("</HEAD>");
  cgiPrintf("<BODY>");
  cgiPrintf("<H2>%i %s</H2>", statusCode, statusReason);
  cgiPrintf("<H3>%s</H3>", statusMesg);
  cgiPrintf("</BODY>");
  cgiPrintf("</HTML>");

  return 1;
}

/******************************************************************************
**
** Function  cgiHeaderNonParse
**
** Purpose   Indicate that the document is to be sent directly to the client
**           with minimal buffering.
**
** History   Date     Author Modification
**           07-03-96 JEM    First Draft
**
******************************************************************************/

int cgiHeaderNonParse(int statusCode, char *statusMesg, char *mimeType)
{
  if (cgiHeaderFlag++)
    return 0;
    
  printf("%s %i %s\n", 
	 *cgiEnv.serverProtocol?cgiEnv.serverProtocol:"unknown", 
	 statusCode, statusMesg);
  printf("Server: %s\n", 
	 *cgiEnv.serverSoftware?cgiEnv.serverSoftware:"unknown");
  printf("Content-type: %s\n\n", mimeType);

  return 1;
}

/******************************************************************************
**
** Function  cgiHtmlHead
**
** Purpose   Essentially a wrapper for the initialisation of a HTML document
**
** History   Date     Author Modification
**           29-03-96 JEM    First Draft
**
******************************************************************************/

int cgiHtmlHead(char *title)
{
  if (!cgiHeaderFlag)
    return 0;

  if (cgiHtmlFlag++)
    return 0;

  cgiPrintf("<HTML>");
  cgiPrintf("<HEAD>");
  cgiPrintf("<TITLE>%s</TITLE>", title);
  cgiPrintf("</HEAD>");
  cgiPrintf("<BODY>");

  return 1;
}

/******************************************************************************
**
** Function  cgiHtmlTail
**
** Purpose   Finalises a HTML document
**
** History   Date     Author Modification
**           29-03-96 JEM    First Draft
**
******************************************************************************/

int cgiHtmlTail()
{
  if (!cgiHeaderFlag)
    return 0;

  if (!cgiHtmlFlag)
    return 0;

  cgiPrintf("</BODY>");
  cgiPrintf("</HTML>");

  return 1;
}

/******************************************************************************
**
** Function  cgiImageOk
**
** Purpose   Determines the browser's ability to view images
**
** History   Date     Author Modification
**           29-03-96 JEM    First Draft
**
******************************************************************************/

int cgiImageOk()
{
  return (!strstr(cgiEnv.httpAccept, "image"));
}

/******************************************************************************
**
** Function  cgiLoad
**
** Purpose   Copy an external file into the current document
**
** History   Date     Author Modification
**           01-04-96 JEM    First Draft
**
******************************************************************************/

int cgiLoad(char *file)
{
  FILE *fp;
  int   c;

  if (!(fp = fopen(file, "r")))
    return 0;

  while ((c=getc(fp)) != EOF)
    putchar(c);

  fflush(stdout);

  fclose(fp);

  return 1;
}

/******************************************************************************
**
** Function  cgiPrintf
**
** Purpose   Same as printf, but suffixes a NL char to help keep script tidy,
**           and flushes output.
**
** History   Date     Author Modification
**           04-03-96 JEM    First Draft
**
******************************************************************************/

int cgiPrintf(char *fmt, ...)
{
  va_list argp;

  if (!cgiHeaderFlag)
    return 0;

  va_start(argp, fmt);
  vfprintf(stdout, fmt, argp);
  va_end(argp);

  fprintf(stdout, "\n");

  fflush(stdout);

  return 1;
}

/******************************************************************************
** Local functions
******************************************************************************/

/******************************************************************************
**
** Function  cgiDebugGetStr
**
** Purpose   Used by cgiDebugEnvLoad to get an environment string
**
** History   Date     Author Modification
**           29-03-96 JEM    First Draft
**           16-04-96 JEM    Return empty string, not NULL, on malloc fail
**
******************************************************************************/

static char *cgiDebugGetStr(FILE *fp)
{
  char *s;
  long  la, lb;

  la = ftell(fp);
  while (getc(fp)!='\n')
    ;
  lb = ftell(fp);

  if (!(s = (char*)malloc(lb-la)))
    return "";

  fseek(fp, la, 0);
  fgets(s, (lb-la), fp);
  fseek(fp, lb, 0);

  return s;
}

/******************************************************************************
**
** Function  cgiDebugGetInt
**
** Purpose   Used by cgiDebugEnvLoad to get an environment integer
**
** History   Date     Author Modification
**           29-03-96 JEM    First Draft
**
******************************************************************************/

static int cgiDebugGetInt(FILE *fp)
{
  char *s = cgiDebugGetStr(fp);
  int   i = atoi(s);

  free(s);
  
  return i;
}

/******************************************************************************
**
** Function  cgiDebugEnvLoad
**
** Purpose   Load a debug environment
**
** History   Date     Author Modification
**           29-03-96 JEM    First Draft
**
******************************************************************************/

static int cgiDebugEnvLoad(char *file)
{
  FILE *fp;
  int   cnt;

  if (!(fp = fopen(file, "rb")))
    return 0;

  /* --- command-line arguments --- */

  cgiArgc = cgiDebugGetInt(fp);

  if (!(cgiArgv = (char**)malloc(cgiArgc*sizeof(char*))))
    return 0;

  for (cnt=0; cnt<cgiArgc; cnt++)
    cgiArgv[cnt] = cgiDebugGetStr(fp);

  /* --- standard input --- */

  cgiStdin = cgiDebugGetStr(fp);

  /* --- CGI environment variables --- */

  cgiEnv.authType         = cgiDebugGetStr(fp);
  cgiEnv.contentLength    = cgiDebugGetInt(fp);
  cgiEnv.contentType      = cgiDebugGetStr(fp);
  cgiEnv.gatewayInterface = cgiDebugGetStr(fp);
  cgiEnv.httpAccept       = cgiDebugGetStr(fp);
  cgiEnv.httpHost         = cgiDebugGetStr(fp);
  cgiEnv.httpReferer      = cgiDebugGetStr(fp);
  cgiEnv.httpUserAgent    = cgiDebugGetStr(fp);
  cgiEnv.pathInfo         = cgiDebugGetStr(fp);
  cgiEnv.pathTranslated   = cgiDebugGetStr(fp);
  cgiEnv.queryString      = cgiDebugGetStr(fp);
  cgiEnv.remoteAddr       = cgiDebugGetStr(fp);
  cgiEnv.remoteHost       = cgiDebugGetStr(fp);
  cgiEnv.remoteIdent      = cgiDebugGetStr(fp);
  cgiEnv.remoteUser       = cgiDebugGetStr(fp);
  cgiEnv.requestMethod    = cgiDebugGetStr(fp);
  cgiEnv.scriptName       = cgiDebugGetStr(fp);
  cgiEnv.serverName       = cgiDebugGetStr(fp);
  cgiEnv.serverPort       = cgiDebugGetStr(fp);
  cgiEnv.serverProtocol   = cgiDebugGetStr(fp);
  cgiEnv.serverSoftware   = cgiDebugGetStr(fp);

  fclose(fp);

  return 1;
}

/******************************************************************************
**
** Function  cgiDebugEnvSave
**
** Purpose   Save a debug environment
**
** History   Date     Author Modification
**           29-03-96 JEM    First Draft
**
******************************************************************************/

static int cgiDebugEnvSave(char *file)
{
  FILE *fp;
  int   cnt;

  if (!(fp = fopen(file, "wb")))
    return 0;

  /* --- command-line arguments --- */

  fprintf(fp, "%i\n", cgiArgc);

  for (cnt=0; cnt<cgiArgc; cnt++)
    fprintf(fp, "%s\n", cgiArgv[cnt]);
    
  /* --- standard input --- */

  fprintf(fp, "%s\n", cgiStdin);

  /* --- CGI environment variables --- */

  fprintf(fp, "%s\n", cgiEnv.authType);
  fprintf(fp, "%i\n", cgiEnv.contentLength);
  fprintf(fp, "%s\n", cgiEnv.contentType);
  fprintf(fp, "%s\n", cgiEnv.gatewayInterface);
  fprintf(fp, "%s\n", cgiEnv.httpAccept);
  fprintf(fp, "%s\n", cgiEnv.httpHost);
  fprintf(fp, "%s\n", cgiEnv.httpReferer);
  fprintf(fp, "%s\n", cgiEnv.httpUserAgent);
  fprintf(fp, "%s\n", cgiEnv.pathInfo);
  fprintf(fp, "%s\n", cgiEnv.pathTranslated);
  fprintf(fp, "%s\n", cgiEnv.queryString);
  fprintf(fp, "%s\n", cgiEnv.remoteAddr);
  fprintf(fp, "%s\n", cgiEnv.remoteHost);
  fprintf(fp, "%s\n", cgiEnv.remoteIdent);
  fprintf(fp, "%s\n", cgiEnv.remoteUser);
  fprintf(fp, "%s\n", cgiEnv.requestMethod);
  fprintf(fp, "%s\n", cgiEnv.scriptName);
  fprintf(fp, "%s\n", cgiEnv.serverName);
  fprintf(fp, "%s\n", cgiEnv.serverPort);
  fprintf(fp, "%s\n", cgiEnv.serverProtocol);
  fprintf(fp, "%s\n", cgiEnv.serverSoftware);
  
  fclose(fp);

  return 1;
}

/******************************************************************************
**
** Function  cgiDecodeUrl
**
** Purpose   Decode a URL-envoded string (plus to space, hex to ascii)
**
** History   Date     Author Modification
**           08-03-96 JEM    First Draft
**           06-12-96 JEM    Bug fix. Due to differing compiler precedence
**                           rules, *++i had to be changed.
**
******************************************************************************/

static char *cgiDecodeUrl(char *data)
{
  char *i, *j;

  for (j=i=data; *i!='\0'; i++, j++)
    {
      if (*i == '%')
	{
	  *j = cgiHexAsc(*(i+1), *(i+2));
	  i += 2;
	}
      else 
      if (*i == '+')
	*j = ' ';
      else
	*j = *i;
    }
  *j = '\0';

  return data;
}

/******************************************************************************
**
** Function  cgiFormAdd
**
** Purpose   Add a form pair to the internal linked list
**
** History   Date     Author Modification
**           15-03-96 JEM    First Draft
**
******************************************************************************/

static int cgiFormAdd(char *var, char *val)
{
  if (!(form.pair))
    form.pair = (pair_t*)malloc((form.size+1)*sizeof(pair_t));
  else
    form.pair = (pair_t*)realloc(form.pair, (form.size+1)*sizeof(pair_t));

  if (!(form.pair))
    return 0;

  if (!(form.pair[form.size].var = strdup(var)) ||
      !(form.pair[form.size].val = strdup(val)))
    return 0;

  ++form.size;

  return 1;
}

/******************************************************************************
**
** Function  cgiGetEnv
**
** Purpose   
**
** History   Date     Author Modification
**           06-03-96 JEM    First Draft
**
******************************************************************************/

static char *cgiGetEnv(char *var)
{
  char *val = getenv(var);

  return (val) ? (val) : "";
}

/******************************************************************************
**
** Function  cgiGetStdin
**
** Purpose   Get CGI data from stdin
**
** History   Date     Author Modification
**           06-03-96 JEM    First Draft
**
******************************************************************************/

static char *cgiGetStdin(int len)
{
  char *data;

  if (!(data = (char*)malloc(len+1)))
    return "";
  
  len = fread(data, 1, len, stdin);

  data[len]='\0';

  return data;
}

/******************************************************************************
**
** Function  cgiHexAsc
**
** Purpose   Convert two hex values to an ascii character
**
** History   Date     Author Modification
**           08-03-96 JEM    First Draft
**
******************************************************************************/

static char cgiHexAsc(char x1, char x2)
{
  return (isxdigit(x1)&&isxdigit(x2))?((cgiHexInt(x1)<<4)+cgiHexInt(x2)):' ';
}

/******************************************************************************
**
** Function  cgiHexInt
**
** Purpose   Convert a hex value to an integer value
**
** History   Date     Author Modification
**           08-03-96 JEM    First Draft
**
******************************************************************************/

static int cgiHexInt(char x)
{
  return (x>='0' && x<='9') ? (x-'0') :
         (x>='a' && x<='f') ? (x-'a'+10) :
         (x>='A' && x<='F') ? (x-'A'+10) : 0;
}

/******************************************************************************
**
** Function  cgiParseArgs
**
** Purpose   Get CGI data from commandline arguments
**
** History   Date     Author Modification
**           06-03-96 JEM    First Draft
**
******************************************************************************/

static int cgiParseArgs(char **argv, int argc)
{
  int cnt;

  for (cnt=1; cnt<argc; cnt++)
    if (!cgiFormAdd("isindex", argv[cnt]))
      return 0;

  return 1;
}

/******************************************************************************
**
** Function  cgiParseInput
**
** Purpose   Parse CGI data, URL-decode and add pairs to linked list
**
** History   Date     Author Modification
**           06-03-96 JEM    First Draft
**
******************************************************************************/

static int cgiParseInput(char *input, int len)
{
  char *var, *val;
  int   cnt, stPnt=0, eqPnt=0, eqCnt=0, amPnt=0, amCnt=0;

  if (!len)
    return 1;

  for (cnt=0; cnt<=len; cnt++)
    {
      if (input[cnt] == '=')
	{
	  eqCnt++;
	  eqPnt = cnt;
	}
      else
      if (input[cnt] == '&' || cnt==len)
	{
	  amCnt++;
	  amPnt = cnt;
	}

      if (!eqCnt && amCnt)    /* no equal in equal/ampersand pair */
	return 0;
      else
      if (eqCnt>1 || amCnt>1) /* equal/ampersand followed by same */
	return 0;
      else
      if (eqCnt && amCnt)
	{
	  if (stPnt == eqPnt) /* not allowed to have a NULL variable */
	    return 0;

	  if (!(var = (char*)calloc(eqPnt-stPnt+1, 1)))
	      return 0;
	  memcpy(var, input+stPnt, eqPnt-stPnt);
	  var = cgiDecodeUrl(var);

	  if(!(val = (char*)calloc(amPnt-eqPnt, 1)))
	     return 0;
	  memcpy(val, input+eqPnt+1, amPnt-eqPnt-1);
	  val = cgiDecodeUrl(val);

	  if (!cgiFormAdd(var, val))
	    return 0;
   
	  free(var);
	  free(val);

	  eqCnt = amCnt = 0;
	  stPnt = amPnt+1;
	}
    }

  return 1;
}

/******************************************************************************
**
** Function  main
**
** Purpose   Main
**
** History   Date     Author Modification
**           06-03-96 JEM    First Draft
**           29-07-96 JEM    Fixed error messages, residue from testing
**
******************************************************************************/

int main(int argc, char **argv)
{
  int cgiStatus = 0;

  if (cgiDebug && strstr(cgiDebug, "LOAD="))
    {
      if (cgiDebugEnvLoad(strchr(cgiDebug, '=')+1))
	{
	  cgiHeaderStatus(200, "Environment Reload Succeeded",
			  "CGI debug environment was loaded from file");
	  cgiPrintf("<HR SIZE=10>");
	}
      else
	{
	  cgiHeaderStatus(505, "Environment Reload Failed",
			  "CGI debug environment failed to reload from file");
	  return 1;
	}
    }
  else
    {
      cgiArgc  = argc;
      cgiArgv  = argv;
      cgiStdin = "";
      
      cgiEnv.authType         = cgiGetEnv("AUTH_TYPE");
      cgiEnv.contentLength    = atoi(cgiGetEnv("CONTENT_LENGTH"));
      cgiEnv.contentType      = cgiGetEnv("CONTENT_TYPE");
      cgiEnv.gatewayInterface = cgiGetEnv("GATEWAY_INTERFACE");
      cgiEnv.httpAccept       = cgiGetEnv("HTTP_ACCEPT");
      cgiEnv.httpHost         = cgiGetEnv("HTTP_HOST");
      cgiEnv.httpReferer      = cgiGetEnv("HTTP_REFERER");
      cgiEnv.httpUserAgent    = cgiGetEnv("HTTP_USER_AGENT");
      cgiEnv.pathInfo         = cgiGetEnv("PATH_INFO");
      cgiEnv.pathTranslated   = cgiGetEnv("PATH_TRANSLATED");
      cgiEnv.queryString      = cgiGetEnv("QUERY_STRING");
      cgiEnv.remoteAddr       = cgiGetEnv("REMOTE_ADDR");
      cgiEnv.remoteHost       = cgiGetEnv("REMOTE_HOST");
      cgiEnv.remoteIdent      = cgiGetEnv("REMOTE_IDENT");
      cgiEnv.remoteUser       = cgiGetEnv("REMOTE_USER");
      cgiEnv.requestMethod    = cgiGetEnv("REQUEST_METHOD");
      cgiEnv.scriptName       = cgiGetEnv("SCRIPT_NAME");
      cgiEnv.serverName       = cgiGetEnv("SERVER_NAME");
      cgiEnv.serverPort       = cgiGetEnv("SERVER_PORT");
      cgiEnv.serverProtocol   = cgiGetEnv("SERVER_PROTOCOL");
      cgiEnv.serverSoftware   = cgiGetEnv("SERVER_SOFTWARE");

      if (!strcmp(cgiEnv.requestMethod, "POST"))
	cgiStdin = cgiGetStdin(cgiEnv.contentLength);
      
      if (!strcmp(cgiEnv.requestMethod, "GET"))
	cgiEnv.contentLength = strlen(cgiEnv.queryString);
    }

  if (!strcmp(cgiEnv.requestMethod, "GET"))
    {
      if (cgiArgc>1)
	{
          if (!cgiParseArgs(cgiArgv, cgiArgc))
            {
              cgiHeaderStatus(405, "Input Not Valid",
			      "Data sent to the CGI program was not valid");
              return 0;
            }
	}
      else
	{
	  if (!cgiParseInput(cgiEnv.queryString, cgiEnv.contentLength))
	    {
	      cgiHeaderStatus(405, "Input Not Valid",
			      "Data sent to the CGI program was not valid");
	      return 1;
	    }
	}
    }
  else
  if (!strcmp(cgiEnv.requestMethod, "POST"))
    {
      if (!strcmp(cgiEnv.contentType, "application/x-www-form-urlencoded"))
	{
	  if (!cgiParseInput(cgiStdin, cgiEnv.contentLength))
	    {
	      cgiHeaderStatus(405, "Input Not Valid",
			      "Data sent to the CGI program was not valid");
	      return 1;
	    }
	}
      else
	{
	  cgiHeaderStatus(406, "Input Not Encoded",
			  "Data send to the CGI program was not URL encoded");
	  return 1;
	}
    }
  else
    {
      cgiHeaderStatus(407, "Unknown Request Method",
		      "Only POST or GET request methods are available");
      return 1;
    }
 
  if (cgiDebug && strstr(cgiDebug, "SAVE="))
    {
      if (cgiDebugEnvSave(strchr(cgiDebug, '=')+1))
	{
	  cgiHeaderStatus(200, "Environment Capture Succeeded",
			  "CGI debug environment was saved to file");
	  cgiPrintf("<HR SIZE=10>");
	}
      else
	{
	  cgiHeaderStatus(505, "Environment Capture Failed",
			  "CGI debug environment was not saved to file");
	  return 1;
	}
    }

  cgiStatus = cgiMain();

  if (!cgiHeaderFlag)
    {
      cgiHeaderStatus(504, "Document Header Unspecified",
		      "Document MIME type was not specified");
      return 1;
    }

  return cgiStatus;
}

/* The End
******************************************************************************/
